import IPython
from IPython.display import Audio, Video, YouTubeVideo, Code


def audio(path: str):
    return IPython.display.display(Audio(path))


def video(path: str, embed=False):
    return IPython.display.display(Video(path, embed=embed))


def youtube_video(id_: str):
    return IPython.display.display(YouTubeVideo(id_))


def code(path: str, language: str = None):
    from pygments import highlight
    from pygments.formatters.html import HtmlFormatter
    # fix JupyterLab from https://github.com/ipython/ipython/issues/11747
    def _jupyterlab_repr_html_(self):
        fmt = HtmlFormatter()
        style = "<style>{}\n{}</style>".format(
            fmt.get_style_defs(".output_html"), fmt.get_style_defs(".jp-RenderedHTML")
        )
        return style + highlight(self.data, self._get_lexer(), fmt)

    IPython.display.Code._repr_html_ = _jupyterlab_repr_html_
    return IPython.display.display(Code(path, language=language))



from . import image_browser
from . import tensorboard
from . import watcher
from .color import Colors, ContinuousColors
from ..backend import VisualLib

__all__ = [
    "watcher",
    "image_browser",
    "tensorboard",
    "Colors",
    "ContinuousColors",
]

if VisualLib.matplotlib_available():
    from . import matplotlib
    from .multi_plots import MultiPlots

    __all__.extend(["matplotlib", "MultiPlots"])

try:
    from . import seaborn
except ImportError:
    pass
else:
    __all__.append("seaborn")



from .event import Event
from .decorator import subscribe, subscribe_async, subscribe_process, subscribe_thread
from .bus import EventBus, EventHandler

__all__ = [
    "Event",
    "EventBus",
    "EventHandler",
    "subscribe",
    "subscribe_async",
    "subscribe_process",
    "subscribe_thread",
]



import pickle


class PickleWriter:

    @classmethod
    def to(cls, path: str, obj: object) -> None:
        """
        Save the object to a file.

        Args:
            path (``str``): The path to the file.
            obj (``object``): The object to save.
        """
        with open(path, 'wb') as f:
            pickle.dump(obj, f)

    def __new__(cls, path: str, obj: object) -> None:
        """Alias to :meth:`~tensorneko_util.io.pickle.PickleWriter.to`."""
        return cls.to(path, obj)



from concurrent.futures import Future
from functools import wraps
from subprocess import Popen
from typing import Callable, Union, Any


def run_blocking(func: Callable[..., Union[Future, Popen]]) -> Callable[..., Any]:
    @wraps(func)
    def wrapper(*args, **kwargs) -> Any:
        task = func(*args, **kwargs)
        if isinstance(task, Future):
            return task.result()
        elif isinstance(task, Popen):
            task.wait()
            return task.returncode
        else:
            raise TypeError("run_blocking only works with Future or Popen")

    return wrapper



import os
import subprocess
from pathlib import Path
from subprocess import Popen
from typing import Optional, List


def ffmpeg_command(ffmpeg_args: List[str] = None, **kwargs) -> Popen:
    """
    Run ffmpeg with the given arguments.

    Args:
        ffmpeg_args (``list``, optional): Additional arguments for ffmpeg.
        **kwargs: Keyword arguments for ~:class:`subprocess.Popen`.

    Returns:
        :class:`subprocess.Popen`: The subprocess of ffmpeg.
    """

    ffmpeg_args = ffmpeg_args or []

    return subprocess.Popen(["ffmpeg"] + ffmpeg_args, **kwargs)


def video2frames(video_path: str, output_dir: Optional[str] = None, frame_name_pattern: str = "%04d.png",
    ffmpeg_args: List[str] = None
) -> Popen:
    """
    Run ffmpeg to extract frames from a video.

    Args:
        video_path (``str``): Path to the video.
        output_dir (``str``, optional): Path to the output directory. If not specified, the output directory will be
            the name of the video. E.g. "path/to/video.mp4" -> "path/to/video".
        frame_name_pattern (``str``, optional): Pattern of the frame name used in ffmpeg command.
        ffmpeg_args (``list``, optional): Additional arguments for ffmpeg.

    Returns:
        :class:`subprocess.Popen`: The subprocess of ffmpeg.
    """

    if output_dir is None:
        path = Path(video_path)
        output_dir = os.path.join(str(path.parent), str("".join(path.name.split(".")[:-1])))

    ffmpeg_args = ffmpeg_args or []

    if not os.path.exists(output_dir):
        Path(output_dir).mkdir(parents=True)

    args = [
        "-i", video_path, os.path.join(output_dir, frame_name_pattern), *ffmpeg_args
    ]

    return ffmpeg_command(args, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)


def frames2video(frames_dir: str, output_path: str, frame_name_pattern: str, fps: float, ffmpeg_args: List[str] = None
) -> Popen:
    """
    Run ffmpeg to convert frames to a video.

    Args:
        frames_dir (``str``): Path to the frames directory.
        output_path (``str``): Path to the output video.
        frame_name_pattern (``str``): Pattern of the frame name used in ffmpeg command.
        fps (``float``): FPS of the output video.
        ffmpeg_args (``list``, optional): Additional arguments for ffmpeg.

    Returns:
        :class:`subprocess.Popen`: The subprocess of ffmpeg.
    """

    ffmpeg_args = ffmpeg_args or []

    args = [
        "-framerate", str(fps), "-i", os.path.join(frames_dir, frame_name_pattern), output_path, *ffmpeg_args
    ]

    return ffmpeg_command(args, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)


def merge_video_audio(video_path: str, audio_path: str, output_path: str, shortest: bool = False,
    ffmpeg_args: List[str] = None
) -> Popen:
    """
    Run ffmpeg to merge video and audio.

    Args:
        video_path (``str``): Path to the video.
        audio_path (``str``): Path to the audio.
        output_path (``str``): Path to the output video.
        shortest (``bool``, optional): If True, the output video will be the shortest possible.
        ffmpeg_args (``list``, optional): Additional arguments for ffmpeg.

    Returns:
        :class:`subprocess.Popen`: The subprocess of ffmpeg.
    """

    ffmpeg_args = ffmpeg_args or []

    if shortest:
        ffmpeg_args.append("-shortest")

    args = [
        "-i", video_path, "-i", audio_path, "-c:v", "copy", "-c:a", "aac", output_path, *ffmpeg_args
    ]

    return ffmpeg_command(args, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)


def resample_video_fps(video_path: str, output_path: str, fps: float, ffmpeg_args: List[str] = None) -> Popen:
    """
    Run ffmpeg to resample video to a specified FPS.

    Args:
        video_path (``str``): Path to the video.
        output_path (``str``): Path to the output video.
        fps (``float``): FPS of the output video.
        ffmpeg_args (``list``, optional): Additional arguments for ffmpeg.

    Returns:
        :class:`subprocess.Popen`: The subprocess of ffmpeg.
    """

    ffmpeg_args = ffmpeg_args or []

    args = [
        "-i", video_path, "-filter:v", f"fps={fps}", output_path, *ffmpeg_args
    ]

    return ffmpeg_command(args, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)


def mp32wav(mp3_path: str, wav_path: str, ffmpeg_args: List[str] = None) -> Popen:
    """
    Run ffmpeg to convert mp3 to wav.

    Args:
        mp3_path (``str``): Path to the mp3.
        wav_path (``str``): Path to the wav.
        ffmpeg_args (``list``, optional): Additional arguments for ffmpeg.

    Returns:
        :class:`subprocess.Popen`: The subprocess of ffmpeg.
    """

    ffmpeg_args = ffmpeg_args or []

    args = [
        "-i", mp3_path, wav_path, *ffmpeg_args
    ]
    return ffmpeg_command(args, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)


def clip_video(video_path: str, output_path: str, start_time: float, end_time: float,
    precise: bool = False,
    ffmpeg_args: List[str] = None
) -> Popen:
    """
    Run ffmpeg to clip video.

    Args:
        video_path (``str``): Path to the video.
        output_path (``str``): Path to the output video.
        start_time (``float``): Start time of the output video.
        end_time (``float``): End time of the output video.
        precise (``bool``, optional): Choose fast or precise. Pass True will use precise mode.
        ffmpeg_args (``list``, optional): Additional arguments for ffmpeg.

    Returns:
        :class:`subprocess.Popen`: The subprocess of ffmpeg.
    """

    ffmpeg_args = ffmpeg_args or []

    if precise:
        args = [
            "-ss", str(start_time), "-to", str(end_time), "-i", video_path, "-c", "copy", output_path, *ffmpeg_args
        ]
    else:
        args = [
            "-i", video_path, "-ss", str(start_time), "-to", str(end_time), "-c", "copy", output_path, *ffmpeg_args
        ]

    return ffmpeg_command(args, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)



import subprocess

from ..util.server import AbstractServer


class Server(AbstractServer):
    def __init__(self, logdir: str = "logs", port: int = 6006):
        self.logdir = logdir
        self.port = port
        self.process = None

    def start(self) -> None:
        # stop the old server
        if self.process is not None:
            self.stop()

        # run new server
        self.process = subprocess.Popen(["tensorboard", "--logdir", self.logdir, "--port", str(self.port)],
            stdout=subprocess.DEVNULL,
            stderr=subprocess.STDOUT
        )
        print(f"Server started at port {self.port}, logdir \"{self.logdir}\".")

    def start_blocking(self) -> None:
        # stop the old server
        if self.process is not None:
            self.stop()

        # run new server
        self.process = subprocess.run(["tensorboard", "--logdir", self.logdir, "--port", str(self.port)])
        print(f"Server started at port {self.port}, logdir \"{self.logdir}\".")

    def stop(self):
        """
        Stop the server.
        """
        if self.process is not None:
            self.process.terminate()
            print("Server stopped")
            self.process = None
        else:
            print("Server is stopped.")



import h5py


class Hdf5Reader:

    @classmethod
    def of(cls, path: str) -> h5py.File:
        """
        Open a hdf5 file.

        Args:
            path (``str``): Path to the hdf5 file.

        Returns:
            :class:`h5py.File`: The opened hdf5 file.

        """
        return h5py.File(path, "r")

    def __new__(cls, path: str):
        return cls.of(path)



from enum import Enum
from typing import TypeVar, Union

# Generic types
from numpy import ndarray

T = TypeVar('T')  # Any type
R = TypeVar('R')  # Return type
P = TypeVar("P", int, float, str, bool)  # Primitive type
E = TypeVar("E", bound=Enum)

K = TypeVar("K")  # Key type
V = TypeVar("V")  # Value type

T1 = TypeVar("T1")
T2 = TypeVar("T2")
T3 = TypeVar("T3")

T_CO = TypeVar("T_CO", covariant=True)  # Covariant type

# array union type
try:
    import torch
except ImportError:
    T_ARRAY = ndarray
else:
    T_ARRAY = Union[ndarray, torch.Tensor]



import pickle


class PickleReader:

    @classmethod
    def of(cls, path: str) -> None:
        """
        Save the object to a file.

        Args:
            path (``str``): The path to the file.
        """
        with open(path, 'rb') as f:
            return pickle.load(f)

    def __new__(cls, path: str) -> None:
        """Alias to :meth:`~tensorneko_util.io.pickle.PickleReader.of`."""
        return cls.of(path)



from .abstract_face_detector import AbstractFaceDetector
from .opencv_face_detector import OpencvFaceDetector
from .facexzoo_face_detector import FaceXZooFaceDetector
from .anime_face_detector import AnimeFaceDetector

__all__ = [
    "AbstractFaceDetector",
    "OpencvFaceDetector",
    "FaceXZooFaceDetector",
    "AnimeFaceDetector"
]



from typing import Dict, Any

import scipy.io


class MatReader:

    @classmethod
    def of(cls, path: str) -> Dict[str, Any]:
        """
        Read a mat file.

        Args:
            path (``str``): The path of the mat file.
        """
        return scipy.io.loadmat(path)

    def __new__(cls, path: str) -> Dict[str, Any]:
        """Alias to :meth:`~tensorneko_util.io.matlab.MatReader.of`."""
        return cls.of(path)



from __future__ import annotations
from typing import TYPE_CHECKING, Tuple, Optional, Union

from ..crop import crop_with_padding

if TYPE_CHECKING:
    from .abstract_face_detector import AbstractFaceDetector


def detector_model_crop_image(detector: AbstractFaceDetector, image_path: str, out_path: str, max_faces=1, margin=0):
    if max_faces > 1:
        raise NotImplementedError("Multiple faces are not supported yet.")
    import cv2

    frame = cv2.imread(image_path)
    dets = detector.detect_face(frame)
    for det in dets[:max_faces]:
        x1, y1, x2, y2, _ = det
        cropped = crop_with_padding(frame, int(x1 - margin), int(x2 + margin), int(y1 - margin), int(y2 + margin))
        cv2.imwrite(out_path, cropped)


def _infer_frame_size(detector: AbstractFaceDetector, video_path: str, margin: int = 0
) -> Tuple[int, int]:
    import cv2
    video = cv2.VideoCapture(video_path)
    while True:
        ret, frame = video.read()
        if not ret:
            break
        dets = detector.detect_face(frame)
        if len(dets) > 0:
            x1, y1, x2, y2, confidence = dets[0]
            # center
            side = int(max(abs(x2 - x1), abs(y2 - y1)))
            video.release()
            return side + 2 * margin, side + 2 * margin


def detector_model_crop_video(detector: AbstractFaceDetector, video_path: str, out_path: str,
    frame_size: Optional[Union[int, Tuple[int, int]]] = None, margin=0, fourcc="mp4v"
):
    import cv2
    video = cv2.VideoCapture(video_path)
    if not video.isOpened():
        raise IOError("Cannot open video file: " + video_path)

    # infer frame size
    if frame_size is None:
        frame_size = _infer_frame_size(detector, video_path, margin)

    if type(frame_size) is int:
        frame_size = frame_size, frame_size

    fps = video.get(cv2.CAP_PROP_FPS)
    writer = cv2.VideoWriter(out_path, cv2.VideoWriter_fourcc(*fourcc), fps, frame_size)
    x1, y1, x2, y2 = 0, 0, 1, 1
    while True:
        ret, frame = video.read()
        if not ret:
            break
        dets = detector.detect_face(frame)
        if len(dets) > 0:
            x1, y1, x2, y2, confidence = dets[0]
            # center
            x, y = int((x1 + x2) / 2), int((y1 + y2) / 2)
            side = int(max(abs(x2 - x1), abs(y2 - y1)))
            x1 = x - side // 2
            x2 = x + side // 2
            y1 = y - side // 2
            y2 = y + side // 2

        cropped = crop_with_padding(frame, int(x1 - margin), int(x2 + margin), int(y1 - margin), int(y2 + margin))
        resized = cv2.resize(cropped, frame_size)
        writer.write(resized)
    video.release()
    writer.release()



from dataclasses import fields, dataclass
from typing import Dict, Any


def json_data(cls):
    """
    The decorator used for JSON parser.

    Examples::

        @json_data
        class Point:
            x: int
            y: int

        # Use for tensorneko read
        point = tensorneko.io.read.text.of_json("path/to/file.json", cls=Point)

        # Use for the Python json loads
        with open("path/to/file.json", "r") as f:
            point = json.loads(f.read(), object_hook=Point)

        print(point.x, point.y)

    """

    def constructor(self, d: Dict[str, Any]):
        for field in fields(self.__class__):
            k = field.name
            t = field.type
            if "typing.List[" in str(t):
                inner_type = t.__args__[0]
                if "typing.List[" in str(inner_type):
                    inner_inner_type = inner_type.__args__[0]
                    try:
                        is_json_data = inner_inner_type.is_json_data
                    except AttributeError:
                        is_json_data = False
                    if is_json_data:
                        v = [[inner_inner_type(e) for e in each] for each in d[k]]
                    else:
                        v = d[k]
                else:
                    try:
                        is_json_data = inner_type.is_json_data
                    except AttributeError:
                        is_json_data = False

                    if is_json_data:
                        v = list(map(inner_type, d[k]))
                    else:
                        v = d[k]
            else:
                try:
                    is_json_data = t.is_json_data
                except AttributeError:
                    is_json_data = False

                if is_json_data:
                    v = t(d[k])
                else:
                    v = d[k]

            self.__dict__[k] = v

    def to_dict(self) -> dict:
        out = {}
        for field in fields(self):
            value = self.__dict__[field.name]
            if "is_json_data" in type(value).__dict__ and type(value).is_json_data:
                value = value.to_dict()
            out[field.name] = value
        return out

    processed_class: type = dataclass(cls)
    processed_class.__init__ = constructor
    processed_class.to_dict = to_dict
    processed_class.is_json_data = True
    return processed_class



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Callable, Optional

from .monad import Monad
from ...type import T_CO as T, R


class Eval(Monad[T], ABC):
    """
    ``Eval`` is a monad that encapsulates the evaluation of a value.

    Example:

        Use `Eval.now` to create an ``Eval`` object that contains the value which is evaluated immediately.

        .. code-block:: python

            from tensorneko.util import Eval

            @Eval.now
            def x():
                print("Eval x")
                return 1  # Eval x

            def f1(value: int) -> int:
                print("Eval f1")
                return value + 1

            y = x.map(f1)  # Eval f1
            print(y.value)  # 2
            print(y.value)  # 2
            print(y.value)  # 2

        Use `Eval.later` to create an ``Eval`` object that contains the value which is evaluated lazily.

        .. code-block:: python

            from tensorneko.util import Eval

            @Eval.later
            def x():
                print("Eval x")
                return 1

            def f1(value: int) -> int:
                print("Eval f1")
                return value + 1

            y = x.map(f1)  # Eval x, Eval f1
            print(y.value)  # 2
            print(y.value)  # 2
            print(y.value)  # 2

        Use `Eval.always` to create an ``Eval`` object that contains the value which will be evaluated when `Eval.value`
            is called.

        .. code-block:: python

            from tensorneko.util import Eval

            @Eval.always
            def x():
                print("Eval x")
                return 1

            def f1(value: int) -> int:
                print("Eval f1")
                return value + 1

            y = x.map(f1)
            print(y.value)  # Eval x, Eval f1, 2
            print(y.value)  # Eval x, Eval f1, 2
            print(y.value)  # Eval x, Eval f1, 2
    """

    def __init__(self, eval_func: Callable[[], T]):
        self._getter = eval_func
        self._value: Optional[T] = None

    @property
    @abstractmethod
    def value(self) -> T:
        ...

    @property
    def evaluated(self) -> bool:
        return self._value is not None

    @classmethod
    def always(cls, eval_func: Callable[[], T]) -> Always[T]:
        return Always(eval_func)

    @classmethod
    def later(cls, eval_func: Callable[[], T]) -> Later[T]:
        return Later(eval_func)

    @classmethod
    def now(cls, eval_func: Callable[[], T]) -> Now[T]:
        return Now(eval_func)

    @classmethod
    def pure(cls, value: T) -> Eval[T]:
        return cls(lambda: value)

    def map(self, f: Callable[[T], R]) -> Eval[R]:
        return self.__class__(lambda: f(self.value))

    def flat_map(self, f: Callable[[T], Eval[R]]) -> Eval[R]:
        return self.__class__(lambda: f(self.value).value)

    def flatten(self: Eval[Eval[T]]) -> Eval[T]:
        return self.value

    def __str__(self):
        return f"{self.__class__.__name__}({self._getter.__name__ if not self.evaluated else self._value})"

    def __repr__(self):
        return self.__str__()


class Always(Eval[T]):
    """Call by name"""

    @property
    def value(self) -> T:
        return self._getter()


class Later(Eval[T]):
    """Call by need"""

    @property
    def value(self) -> T:
        if self._value is None:
            self._value = self._getter()
        return self._value


class Now(Eval[T]):
    """Call by value"""

    def __init__(self, eval_func: Callable[[], T]):
        super().__init__(eval_func)
        self._value = self._getter()

    @property
    def value(self) -> T:
        return self._value



from argparse import ArgumentParser
from typing import Any


class Arguments:
    def __init__(self):
        self._fields = {}

    def add(self, name, value):
        self._fields[name] = value
        self.__dict__[name] = value

    def __getitem__(self, item):
        return self._fields[item]

    def __setitem__(self, key, value):
        self._fields[key] = value
        self.__dict__[key] = value

    def __str__(self):
        return f"Args {self._fields} "

    def __repr__(self):
        return self.__str__() if len(self._fields) < 10 else "Args (\n    " + \
            "\n    ".join([f"{k}={v}" for k, v in self._fields.items()]) + "\n)"


def get_parser_default_args(parser: ArgumentParser, **kwargs) -> Any:
    args = Arguments()
    for name, value in kwargs.items():
        args.add(name, value)
    for action in parser._actions:
        if action.dest not in ("help", "version") and action.dest not in args._fields:
            args.add(action.dest, action.default)
    return args



from __future__ import annotations

import asyncio
import warnings
from abc import ABC, abstractmethod
from concurrent.futures import wait
from dataclasses import dataclass
from enum import Enum
from typing import Dict, Type, Callable, TYPE_CHECKING, Any, Coroutine, Union, Set, TypeVar

from ...backend.parallel import ExecutorPool, ParallelType

if TYPE_CHECKING:
    from .event import E


class EmitType(Enum):
    NORMAL = 0
    THREAD = 1
    PROCESS = 2
    ASYNC = 3


@dataclass
class Listener:
    func: Callable[[E], Union[None, Coroutine[Any, Any, None]]]
    func_name: str
    emit_type: EmitType

    func_names = set()

    def __post_init__(self):
        if self.func.__name__ in self.func_names and self.emit_type == EmitType.PROCESS:
            warnings.warn(f"Multiprocess event handler {self.func_name} is already registered")
        self.func_names.add(self.func_name)

    def __hash__(self):
        return hash(self.func) + hash(self.emit_type)


class EventHandler(ABC):
    """
    EventHandler is an abstract class that is used to define stateful event handlers.

    The `__init__` method should not have any arguments.

    The `__call__` method should include only one argument `event: EventType` and will be invoked when an event is
        emitted.

    Examples::

        from tensorneko.util import Event, EventHandler, subscribe

        class AddValueEvent(Event):
            def __init__(self, value: int):
                self.value = value

        @subscribe
        class CustomHandler(EventHandler):
            def __init__(self):
                self.x = 0

            def __call__(self, event: AddValueEvent):
                self.x += event.value

        if __name__ == '__main__':
            AddValueEvent(42)
            assert CustomHandler.instance().x == 42

    """

    _instance = None

    @abstractmethod
    def __call__(self, event: E):
        pass

    @classmethod
    def instance(cls):
        if cls._instance is None:
            cls._instance = cls()
        return cls._instance


H = TypeVar("H", bound=EventHandler)


class EventBus:
    """
    Event bus can be used to program the reactive behavior of the system. A default bus is provided as `bus`, and it
    can be imported as `from tensorneko.util import bus`.

    The stateless event handler can be registered to the default bus by using the `@subscribe` decorator.

    Examples:

        Use default event bus.

        .. code-block:: python

            # useful decorators for default event bus
            from tensorneko.util import (
                subscribe, # run in the main thread
                subscribe_thread, # run in a new thread
                subscribe_async, # run async
                subscribe_process # run in a new process
            )
            # Event base type
            from tensorneko.util import Event

            class LogEvent(Event):
                def __init__(self, message: str):
                    self.message = message

            # the event argument should be annotated correctly
            @subscribe
            def log_information(event: LogEvent):
                print(event.message)

            @subscribe_thread
            def log_information_thread(event: LogEvent):
                print(event.message, "in another thread")

            @subscribe_async
            async def log_information_async(event: LogEvent):
                print(event.message, "async")

            @subscribe_process
            def log_information_process(event: LogEvent):
                print(event.message, "in a new process")

            if __name__ == '__main__':
                # emit an event, and then the event handler will be invoked
                # The sequential order is not guaranteed
                LogEvent("Hello world!")

                # one possible output:
                # Hello world! in another thread
                # Hello world! async
                # Hello world!
                # Hello world! in a new process

        Create multiple event bus

        .. code-block:: python

            from tensorneko.util import EventBus
            bus1 = EventBus()
            bus2 = EventBus()

            # register event handlers to different event bus
            @bus1.subscribe
            def log_information(event: LogEvent):
                print(event.message, "in bus1")

            @bus2.subscribe
            def log_information(event: LogEvent):
                print(event.message, "in bus2")

            if __name__ == '__main__':
                # emit the event with specified bus
                LogEvent("Hello world!", bus=bus1)
                LogEvent("Hello world!", bus=bus2)

                # one possible output:
                # Hello world! in bus1
                # Hello world! in bus2
    """

    default: EventBus = None

    def __init__(self):
        self.listeners: Dict[Type[E], Set[Listener]] = {}

    def register(self,
        event_type: Type[E],
        func: Callable[[E], None],
        func_name: str = None,
        emit_type: EmitType = EmitType.NORMAL
    ) -> None:
        if func_name is None:
            func_name = func.__name__

        if event_type not in self.listeners:
            self.listeners[event_type] = set()
        self.listeners[event_type].add(Listener(func, func_name, emit_type))

    def emit(self, event: E, blocking: bool = True) -> None:
        event_type: Type[E] = type(event)
        if event_type in self.listeners:
            normal_listeners = set()
            thread_listeners = set()
            process_listeners = set()
            async_listeners = set()

            for listener in self.listeners[event_type]:
                if listener.emit_type == EmitType.NORMAL:
                    normal_listeners.add(listener)
                elif listener.emit_type == EmitType.THREAD:
                    thread_listeners.add(listener)
                elif listener.emit_type == EmitType.PROCESS:
                    process_listeners.add(listener)
                elif listener.emit_type == EmitType.ASYNC:
                    async_listeners.add(listener)
                else:
                    raise ValueError(f"Unknown emit type: {listener.emit_type}")

            # priority: Process > Thread > Async > Normal
            futures = []
            for listener in process_listeners:
                futures.append(ExecutorPool.submit(listener.func, event, parallel_type=ParallelType.PROCESS))
            for listener in thread_listeners:
                futures.append(ExecutorPool.submit(listener.func, event, parallel_type=ParallelType.THREAD))
            if len(async_listeners) > 0:
                asyncio.run(self.emit_async(async_listeners, event))
            for listener in normal_listeners:
                listener.func(event)

            if blocking:
                wait(futures)

    @staticmethod
    async def emit_async(async_listeners: Set[Listener], event: E) -> None:
        coroutines = []
        for listener in async_listeners:
            coroutines.append(asyncio.create_task(listener.func(event)))
        for coroutine in coroutines:
            await coroutine

    def subscribe(self,
        func: Union[Callable[[E], Union[None, Coroutine[Any, Any, None]]], Type[H]],
        emit_type: EmitType = EmitType.NORMAL
    ) -> Callable[[E], None]:
        if isinstance(func, type):
            event_cls = func
            if not issubclass(event_cls, EventHandler):
                raise ValueError(f"{event_cls.__name__} is not inherited from EventHandler")

            event_handler = event_cls.instance()
            func = event_handler.__call__
            func_name = event_handler.__class__.__name__
            return_value = event_cls
        else:
            func_name = func.__name__
            return_value = func

        if "event" not in func.__annotations__:
            raise TypeError("@subscribe decorator must follow the signature `handler_func(event: EventType)`. "
                            "The event type should be annotated.")

        event_type = func.__annotations__["event"]
        self.register(event_type, func, func_name, emit_type)

        return return_value

    def subscribe_async(self, func: Callable[[E], Coroutine[Any, Any, None]]):
        return self.subscribe(func, emit_type=EmitType.ASYNC)

    def subscribe_thread(self, func: Callable[[E], None]):
        return self.subscribe(func, emit_type=EmitType.THREAD)

    def subscribe_process(self, func: Callable[[E], None]):
        return self.subscribe(func, emit_type=EmitType.PROCESS)


EventBus.default = EventBus()



from typing import Optional

from .audio_data import AudioData
from .._default_backends import _default_audio_io_backend
from ...backend.audio_lib import AudioLib


class AudioReader:
    """AudioReader for reading audio file"""

    @staticmethod
    def of(path: str, channel_first: bool = True, backend: Optional[AudioLib] = None) -> AudioData:
        """
        Read audio tensor from given file.

        Args:
            path (``str``): Path to the audio file.
            channel_first (``bool``, optional): Whether the audio is channel first. The output shape is (C, T) if true
                and (T, C) if false. Default: True.
            backend (:class:`~tensorneko.io.audio.audio_lib.AudioLib`, optional): The audio library to use.
                Default: pytorch.

        Returns:
            :class:`~.audio_data.AudioData`: The Audio data in the file.
        """
        backend = backend or _default_audio_io_backend()

        if backend == AudioLib.PYTORCH:
            if not AudioLib.pytorch_available():
                raise ValueError("Torchaudio is not available.")
            import torchaudio
            return AudioData(*torchaudio.load(path, channels_first=channel_first))
        else:
            raise ValueError("Unknown audio library: {}".format(backend))

    def __new__(cls, path: str, channel_first: bool = True, backend: Optional[AudioLib] = None) -> AudioData:
        """Alias of :meth:`~AudioReader.of`"""
        return cls.of(path, channel_first, backend)



from typing import Type

from .audio import AudioWriter
from .image import ImageWriter
from .json import JsonWriter
from .pickle import PickleWriter
from .text import TextWriter
from .video import VideoWriter

try:
    from .matlab import MatWriter
    scipy_available = True
except ImportError:
    scipy_available = False
    MatWriter = None

try:
    from .yaml import YamlWriter
    yaml_available = True
except ImportError:
    YamlWriter = object
    yaml_available = False

try:
    from .hdf5 import Hdf5Writer
    h5py_available = True
except ImportError:
    Hdf5Writer = object
    h5py_available = False


class Writer:

    def __init__(self):
        self.image = ImageWriter
        self.text = TextWriter
        self.json = JsonWriter
        self.video = VideoWriter
        self.audio = AudioWriter
        self.pickle = PickleWriter
        self._mat = None
        self._yaml = None
        self._h5 = None

    @property
    def mat(self) -> Type[MatWriter]:
        if scipy_available:
            if self._mat is None:
                self._mat = MatWriter
            return self._mat
        else:
            raise ImportError("Scipy is not available to write mat file.")

    @property
    def yaml(self) -> Type[YamlWriter]:
        if yaml_available:
            if self._yaml is None:
                self._yaml = YamlWriter
            return self._yaml
        else:
            raise ImportError("To use the yaml writer, please install pyyaml.")

    @property
    def h5(self) -> Type[Hdf5Writer]:
        if h5py_available:
            if self._h5 is None:
                self._h5 = Hdf5Writer
            return self._h5
        else:
            raise ImportError("To use the hdf5 writer, please install h5py.")



from pathlib import Path
from typing import Optional
from urllib.request import urlretrieve

from ..backend._tqdm import import_tqdm_auto

try:
    auto = import_tqdm_auto()
    tqdm = auto.tqdm
    _is_progress_bar_available = True
except ImportError:
    _is_progress_bar_available = False
    DownloadProgressBar = object
else:
    class DownloadProgressBar(tqdm):
        total: int

        def update_to(self, b=1, bsize=1, tsize=None):
            if tsize is not None:
                self.total = tsize
            self.update(b * bsize - self.n)


def download_file(url: str, dir_path: str = ".", file_path: Optional[str] = None, progress_bar: bool = True) -> str:
    """
    Download file with given URL to given directory path with progress bar.

    Args:
        url (``str``): URL of the file to download.
        dir_path (``str``, optional): Directory path to download the file to. The saved name is the same as the
            original URL name. Defaults to current directory.
        file_path (``str``, optional): File path to download the file to (override dir_path parameter).
            Default None, which uses the dir_path argument.
        progress_bar (``bool``, optional): Whether to show progress bar. Defaults to True.

    Returns:
        ``str``: File path of the downloaded file.

    """
    if file_path is not None:
        path = Path(file_path)
    else:
        path = Path(dir_path) / url.split("/")[-1]
    path.parent.mkdir(exist_ok=True, parents=True)

    if progress_bar:
        if not _is_progress_bar_available:
            raise ImportError("Please install tqdm to use progress bar.")

        with DownloadProgressBar(unit="B", unit_scale=True, miniters=1, desc=f"Downloading {path.name}") as pb:
            urlretrieve(url, filename=path, reporthook=pb.update_to)
    else:
        urlretrieve(url, filename=path)

    return str(path)



from .video import frames2video
from .image import rgb2gray, rgb2gray_batch
from .crop import crop_with_padding
from . import face_detector
from ..backend import VisualLib

__all__ = ['frames2video', 'rgb2gray', 'rgb2gray_batch', 'crop_with_padding', 'face_detector']

if VisualLib.ffmpeg_available():
    from .ffmpeg import ffmpeg_command, video2frames, merge_video_audio, resample_video_fps, mp32wav

    __all__ += ['ffmpeg_command', 'video2frames', 'merge_video_audio', 'resample_video_fps', 'mp32wav']



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import List

import http.server
import os
import shutil
import subprocess
from http.server import HTTPServer
from pathlib import Path
from threading import Thread
from typing import Optional, Union, List


class AbstractServer(ABC):

    @abstractmethod
    def start(self):
        ...

    @abstractmethod
    def stop(self):
        ...

    def __enter__(self) -> AbstractServer:
        self.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.stop()


class HttpThreadServer(AbstractServer, ABC):
    servers: List[HttpThreadServer] = []

    def __init__(self, path: str, port: int = 8000, name: str = "tensorneko server"):
        self.server_dir = path
        self.port = port
        self.name = name
        self.process: Optional[Thread] = None
        self.httpd: Optional[HTTPServer] = None

        class HTTPHandler(http.server.SimpleHTTPRequestHandler):

            def __init__(self, *args, **kwargs):
                super().__init__(*args, directory=path, **kwargs)

            def log_message(self, format: str, *args) -> None:
                return

        self.Handler = HTTPHandler
        self.Handler.extensions_map[".js"] = "application/javascript"
        self.Handler.extensions_map[".mjs"] = "application/javascript"
        HTTPServer.allow_reuse_address = True
        HTTPServer.allow_reuse_port = True
        self.servers.append(self)

    def start(self) -> None:
        """
        Start or restart the server.
        """
        # stop the old server
        if self.process is not None:
            self.stop()
        # copy files to the view directory
        self._prepare()
        # run python http server
        self._run()

    def start_blocking(self) -> None:
        """
        Start or restart the server on blocking.
        """
        # stop the old server
        if self.process is not None:
            self.stop()
        # copy files to the view directory
        self._prepare()
        # run python http server
        self._run_blocking()

    @abstractmethod
    def _prepare(self) -> None:
        pass

    def _run(self) -> None:
        self.httpd = http.server.HTTPServer(("", self.port), self.Handler)
        self.process = Thread(target=self.httpd.serve_forever)
        self.process.start()
        print(self.server_start_string)

    def _run_blocking(self):
        with http.server.HTTPServer(("", self.port), self.Handler) as httpd:
            print(self.server_start_string)
            httpd.serve_forever()

    def stop(self) -> None:
        """
        Stop the server.
        """
        if self.process is not None:
            self.httpd.shutdown()
            self.process.join()
            print("Server stopped")
            self.process = None
        else:
            print("Server is stopped.")

    @classmethod
    def stop_all(cls) -> None:
        """
        Stop all servers.
        """
        for server in cls.servers:
            server.stop()

    @property
    def server_start_string(self) -> str:
        return f"[Server] {self.name} started at http://127.0.0.1:{self.port}/."



import subprocess
from enum import Enum
from typing import Optional


class VisualLib(Enum):
    OPENCV = 1
    PYTORCH = 2
    MATPLOTLIB = 3
    PIL = 4
    FFMPEG = 5
    SKIMAGE = 6

    @staticmethod
    def opencv_available() -> bool:
        if _VisualLibAvailability.is_opencv_available is None:
            try:
                import cv2
                _VisualLibAvailability.is_opencv_available = True
            except ImportError:
                _VisualLibAvailability.is_opencv_available = False
        return _VisualLibAvailability.is_opencv_available

    @staticmethod
    def pytorch_available() -> bool:
        if _VisualLibAvailability.is_torchvision_available is None:
            try:
                import torchvision
                _VisualLibAvailability.is_torchvision_available = True
            except ImportError:
                _VisualLibAvailability.is_torchvision_available = False
        return _VisualLibAvailability.is_torchvision_available

    @staticmethod
    def matplotlib_available() -> bool:
        if _VisualLibAvailability.is_matplotlib_available is None:
            try:
                import matplotlib
                _VisualLibAvailability.is_matplotlib_available = True
            except ImportError:
                _VisualLibAvailability.is_matplotlib_available = False
        return _VisualLibAvailability.is_matplotlib_available

    @staticmethod
    def pil_available() -> bool:
        if _VisualLibAvailability.is_pil_available is None:
            try:
                import PIL
                _VisualLibAvailability.is_pil_available = True
            except ImportError:
                _VisualLibAvailability.is_pil_available = False
        return _VisualLibAvailability.is_pil_available

    @staticmethod
    def ffmpeg_available() -> bool:
        if _VisualLibAvailability.is_ffmpeg_available is None:
            ffmpeg_available = subprocess.run('ffmpeg -version', stdout=subprocess.DEVNULL,
                stderr=subprocess.DEVNULL, shell=True).returncode == 0
            _VisualLibAvailability.is_ffmpeg_available = ffmpeg_available
        return _VisualLibAvailability.is_ffmpeg_available

    @staticmethod
    def skimage_available() -> bool:
        if _VisualLibAvailability.is_skimage_available is None:
            try:
                import skimage
                _VisualLibAvailability.is_skimage_available = True
            except ImportError:
                _VisualLibAvailability.is_skimage_available = False
        return _VisualLibAvailability.is_skimage_available


class _VisualLibAvailability:
    is_opencv_available: Optional[bool] = None
    is_torchvision_available: Optional[bool] = None
    is_matplotlib_available: Optional[bool] = None
    is_pil_available: Optional[bool] = None
    is_ffmpeg_available: Optional[bool] = None
    is_skimage_available: Optional[bool] = None



from ..backend.visual_lib import VisualLib


def _default_image_io_backend():
    if VisualLib.opencv_available():
        backend = VisualLib.OPENCV
    elif VisualLib.matplotlib_available():
        backend = VisualLib.MATPLOTLIB
    else:
        raise ValueError("No image reader backend available.")
    return backend


def _default_video_io_backend():
    if VisualLib.opencv_available():
        return VisualLib.OPENCV
    elif VisualLib.pytorch_available():
        return VisualLib.PYTORCH
    elif VisualLib.ffmpeg_available():
        return VisualLib.FFMPEG
    else:
        raise ValueError("No backend available. Please install OpenCV, Torchvision or FFMPEG.")


def _default_audio_io_backend():
    if VisualLib.pytorch_available():
        return VisualLib.PYTORCH
    else:
        raise ValueError("No backend available. Please install Torchaudio.")



from __future__ import annotations

import os.path
from abc import ABC, abstractmethod
from typing import Generic, List, Union, Dict, TYPE_CHECKING, Any, Optional

from einops import rearrange

from ...backend import VisualLib
from ...util.ref import Ref
from ...util.type import T, P, T_ARRAY

if TYPE_CHECKING:
    from .view import View


class Component(ABC, Generic[T]):
    name: str
    views: List[View]
    _value: T
    components: Dict[str, Component] = {}

    def __init__(self, name: str, value: T):
        self.name = name
        self._value = value
        self.views = []
        Component.components[self.name] = self

    @property
    def value(self) -> T:
        return self._value

    @value.setter
    def value(self, value_in: T) -> None:
        self._value = value_in
        self.update_view()

    def update_view(self):
        for view in self.views:
            view.update()

    def set(self, value_in: T) -> None:
        self.value = value_in

    @abstractmethod
    def to_dict(self) -> Dict[str, Any]:
        ...

    def update(self) -> None:
        pass

    def __str__(self) -> str:
        return f"<{self.__class__.__name__}: {self.to_dict()}>"

    def __repr__(self) -> str:
        return self.__str__()

    def bind(self, ref: Ref[P]) -> Bindable[P]:
        ...


class Bindable(ABC, Generic[P]):
    ref: Ref[P]

    @property
    def value(self):
        return self.ref.value

    @value.setter
    def value(self, value_in: P):
        raise ValueError("Cannot set value of bindable component")


class Variable(Component[T]):
    """
    The components containing a value.

    Args:
        name (``str``): The name of the variable, which should be unique in
            a :class:`~tensorneko.visualization.watcher.view.View`.
        value (``T``): The wrapped value for watching.

    Attributes:
        views (``List`` [:class:`~tensorneko.visualization.web.view.View`]): The views that the variable belongs to.

    Examples::

        # create a Variable component
        var = tensorneko.visualization.watcher.Variable("x", 5)
        # update value
        var.value = 10
        # equivalent to
        var.set(15)

    """

    def __init__(self, name: str, value: T):
        super().__init__(name, value)

    def to_dict(self) -> Dict[str, Any]:
        return {
            "type": "Variable",
            "name": self.name,
            "value": str(self.value)
        }

    def bind(self, ref: Ref[P]) -> BindableVariable[P]:
        return BindableVariable(ref, self.name)


class BindableVariable(Bindable[P], Variable[P]):

    def __init__(self, ref: Ref[P], name: str):
        super().__init__(name, ref.value)
        Bindable.__init__(self)
        self.ref = ref
        self.ref.bound_comp = self


class ProgressBar(Component[int]):
    """
    The components as a progress bar.

    Args:
        name (``str``):
            The name of the variable, which should be unique in a :class:`~tensorneko.visualization.watcher.view.View`.
        value (``int``):
            The current progress of the progress bar.
        total (``int``):
            The maximum number of the progress bar.

    Attributes:
        views (``List`` [:class:`~tensorneko.visualization.web.view.View`]): The views that the variable belongs to.

    Examples::

        # create a ProgressBar component
        pb = tensorneko.visualization.watcher.ProgressBar("process", 0, 10000)
        # update value. these 4 ways below are all equivalent.
        pb.value += 1
        pb.value = pb.value + 1
        pb.set(pb.value + 1)
        pb.add(1)

    """
    total: int

    def __init__(self, name: str, total: int, value: int = 0):
        super().__init__(name, value)
        self.total = total

    def add(self, value_in: int) -> None:
        self.value += value_in

    def to_dict(self) -> Dict[str, Any]:
        return {
            "type": "ProgressBar",
            "name": self.name,
            "value": self.value,
            "total": self.total
        }

    def bind(self, ref: Ref[P]) -> BindableProgressBar[P]:
        return BindableProgressBar(ref, self.name, self.total)


class BindableProgressBar(Bindable[int], ProgressBar):

    def __init__(self, ref: Ref[P], name: str, total: int):
        super().__init__(name, total, ref.value)
        Bindable.__init__(self)
        self.ref = ref
        self.ref.bound_comp = self


class Image(Component[Optional[T_ARRAY]]):
    """
    The component of image for display.

    Args:
        name (``str``): The name of the image for display, which should be unique in
            a :class:`~tensorneko.visualization.watcher.view.View`.

        value (:class:`~numpy.ndarray` | :class:`~torch.Tensor` | ``None``, optional): The image array with (C, H, W).
            Value range are between [0, 1].

    Attributes:
        path (``str``): The image path related to the :class:`~tensorneko.visualization.watcher.view.View` root dir.

    Examples::

        # create an Image component
        img_comp = tensorneko.visualization.watcher.Image("img0", img_arr)
        # update image. these 2 ways below are all equivalent.
        img_comp.value = new_img_arr
        img_comp.set(new_img_arr)

    """

    def __init__(self, name: str, value: Optional[T_ARRAY] = None):
        super().__init__(name, value)
        self.path = os.path.join("img", self.name)
        self._ver = 0
        if VisualLib.matplotlib_available():
            import matplotlib.pyplot as plt
            self._plt = plt
        else:
            raise ImportError("matplotlib is required for Image component.")

    def to_dict(self) -> Dict[str, Any]:
        self.update()
        return {
            "type": "Image",
            "name": self.name,
            "value": f"/img/{self.name}-{self._ver}.jpg"
        }

    def update(self) -> None:
        if self.value is not None:
            for view in self.views:
                img_dir = os.path.join("watcher", view.name, "img")
                if not os.path.exists(img_dir):
                    os.mkdir(img_dir)

                # remove old image
                pre_img_path = os.path.join("watcher", view.name, self.path) + f"-{self._ver}.jpg"
                if os.path.exists(pre_img_path):
                    os.remove(pre_img_path)

                # save new image
                self._ver += 1
                img_path = os.path.join("watcher", view.name, self.path) + f"-{self._ver}.jpg"
                # C, H, W
                image = rearrange(self.value, "c h w -> h w c")
                self._plt.imsave(img_path, image)


class Logger(Component[List[str]]):
    """
    The component of log texts for display.

    Args:
        name (``str``): The name of the image for display, which should be unique in
            a :class:`~tensorneko.visualization.watcher.view.View`.

        value (``List[str]``, optional): The initial logs. Default [].

    Examples::

        # create a Logger component
        logger = tensorneko.visualization.watcher.Logger("training_log")
        # log new message
        logger.log("Epoch 1, loss: 0.1234, val_loss: 0.2345")

    """

    def __init__(self, name: str, value: List[str] = None):
        value = value or []
        super().__init__(name, value)

    def log(self, msg: str) -> None:
        self._value.append(msg)
        self.update_view()

    def to_dict(self) -> Dict[str, Any]:
        return {
            "type": "Logger",
            "name": self.name,
            "value": self.value
        }


class LineChart(Component[List[Dict[str, Union[float, str]]]]):
    """
    The component for logging a line chart.
    """

    def __init__(self, name: str, value: Optional[List[Dict[str, Union[float, str]]]] = None,
        x_label: str = "index", y_label: str = "value"
    ):
        value = value or []
        super().__init__(name, value)
        self.x_label = x_label
        self.y_label = y_label

    def add(self, x: float, y: float, label: str = "") -> None:
        self._value.append({"x": x, "y": y, "label": label})
        for view in self.views:
            view.update()

    def to_dict(self) -> Dict[str, Any]:
        return {
            "type": "LineChart",
            "name": self.name,
            "x_label": self.x_label,
            "y_label": self.y_label,
            "value": self.value
        }



from typing import List, Optional, Union, Tuple

from numpy import ndarray

from ._utils import detector_model_crop_image, detector_model_crop_video
from .abstract_face_detector import AbstractFaceDetector


class AnimeFaceDetector(AbstractFaceDetector):
    """
    hysts/anime-face-detector
    https://github.com/hysts/anime-face-detector
    """

    def __init__(self, device: str = "cuda:0", box_scale_factor: float = 1.1):
        super().__init__()
        try:
            from anime_face_detector import create_detector
            import cv2
        except ImportError:
            raise ImportError("anime-face-detector is not installed. Please install it first. Please check the "
                              "details in https://github.com/hysts/anime-face-detector")
        else:
            self.cv2 = cv2
            self.detector = create_detector(device=device, box_scale_factor=box_scale_factor)

    def detect_face(self, image: ndarray, *args, **kwargs) -> List[ndarray]:
        assert image.ndim == 3 and image.shape[2] == 3, "frame should be 3-dim"
        return [each["bbox"] for each in self.detector.detect(image)]

    def crop_image(self, image_path: str, out_path: str, max_faces=1, margin=0, *args, **kwargs) -> None:
        detector_model_crop_image(self, image_path, out_path, max_faces, margin)

    def crop_video(self, video_path: str, out_path: str, frame_size: Optional[Union[int, Tuple[int, int]]] = None,
        margin=0, fourcc="mp4v", *args, **kwargs
    ) -> None:
        detector_model_crop_video(self, video_path, out_path, frame_size, margin, fourcc)




from dataclasses import dataclass
from typing import Optional

from ...util.type import T_ARRAY


@dataclass
class VideoInfo:
    """A dataclass for info of video"""
    video_fps: float
    audio_fps: Optional[int]

    def __iter__(self):
        return iter((self.video_fps, self.audio_fps))


@dataclass
class VideoData:
    """
    A dataclass for output of :class:`~.video_reader.VideoReader`

    Attributes:
        video (:class:`~torch.Tensor` | :class:`~numpy.ndarray`): video float tensor of (T, C, H, W)
        audio (:class:`~torch.Tensor` | :class:`~numpy.ndarray`): audio float tensor of (C, T)
        info (:class:`VideoInfo`): video and audio fps info
    """
    video: T_ARRAY
    audio: T_ARRAY
    info: VideoInfo

    def __init__(self, video, audio, info):
        self.video = video
        self.audio = audio
        info_args = info if len(info) == 2 else {
            "video_fps": info["video_fps"],
            "audio_fps": None
        }
        self.info = VideoInfo(**info_args)

    def __iter__(self):
        return iter((self.video, self.audio, self.info))



from .reader import Reader
from .writer import Writer

from .json import json_data
from ..backend.visual_lib import VisualLib
from ..backend.audio_lib import AudioLib

read = Reader()
write = Writer()

__all__ = [
    "read",
    "write",
    "json_data",
    "VisualLib",
    "AudioLib",
]



import h5py


class Hdf5Writer:

    @classmethod
    def to(cls, path: str):
        raise NotImplementedError("Hdf5Writer is not implemented yet.")

    def __new__(cls, *args, **kwargs):
        raise NotImplementedError("Hdf5Writer is not implemented yet.")



import os.path
from os.path import dirname

from . import notebook
from . import util
from . import visualization
from . import preprocess
from . import io
from . import debug

__all__ = [
    "notebook",
    "util",
    "visualization",
    "preprocess",
    "io",
    "debug"
]

with open(os.path.join(dirname(__file__), "version.txt"), "r", encoding="UTF-8") as file:
    __version__ = file.read()



from typing import Type

from .audio import AudioReader
from .image import ImageReader
from .json import JsonReader
from .pickle import PickleReader
from .text import TextReader
from .video import VideoReader

try:
    from .matlab import MatReader
    scipy_available = True
except ImportError:
    scipy_available = False
    MatReader = None

try:
    from .yaml import YamlReader
    yaml_available = True
except ImportError:
    YamlReader = object
    yaml_available = False

try:
    from .hdf5 import Hdf5Reader
    h5py_available = True
except ImportError:
    Hdf5Reader = object
    h5py_available = False


class Reader:

    def __init__(self):
        self.image = ImageReader
        self.text = TextReader
        self.json = JsonReader
        self.video = VideoReader
        self.audio = AudioReader
        self.pickle = PickleReader
        self._mat = None
        self._yaml = None
        self._h5 = None

    @property
    def mat(self) -> Type[MatReader]:
        if scipy_available:
            if self._mat is None:
                self._mat = MatReader
            return self._mat
        else:
            raise ImportError("Scipy is not available to read mat file.")

    @property
    def yaml(self) -> Type[YamlReader]:
        if yaml_available:
            if self._yaml is None:
                self._yaml = YamlReader
            return self._yaml
        else:
            raise ImportError("To use the yaml reader, please install pyyaml.")

    @property
    def h5(self) -> Type[Hdf5Reader]:
        if h5py_available:
            if self._h5 is None:
                self._h5 = Hdf5Reader
            return self._h5
        else:
            raise ImportError("To use the hdf5 reader, please install h5py.")



import os.path
import sys
from typing import Optional, Union, Tuple

from numpy import ndarray

from ._utils import detector_model_crop_image, detector_model_crop_video
from .abstract_face_detector import AbstractFaceDetector
from ...io import read


class FaceXZooFaceDetector(AbstractFaceDetector):
    """
    The face sdk is in https://github.com/ControlNet/FaceX-Zoo.
    The fork of https://github.com/JDAI-CV/FaceX-Zoo/tree/main/face_sdk.

    Args:
        face_sdk_path (``str``): The path to the face_sdk path of FaceXZoo.
        device (``str``, optional): The device to use, default is "cuda:0".

    Examples::

        from tensorneko_util.preprocess.face_detector import FaceXZooFaceDetector

        detector = FaceXZooFaceDetector(FaceXZooFaceDetector.install(), "cuda:0")

        # process image, save to facexzoo_face_detector_out.jpg
        detector.process_image("test/resource/image_sample/1.183.jpg",
            "test/resource/temp_output/facexzoo_face_detector_out.jpg")

        # process video, save to facexzoo_face_detector_out.mp4
        detector.process_video("test/resource/video_sample/sample.mp4",
            "test/resource/temp_output/facexzoo_face_detector_out.mp4")

    """

    def __init__(self, face_sdk_path: Optional[str] = None, device: str = "cuda:0"):
        try:
            import cv2
        except ImportError:
            raise ImportError("OpenCV and opencv-contrib-python is not installed. Please install it first.")

        self.cv2 = cv2

        if face_sdk_path is not None:
            sys.path.append(face_sdk_path)
        else:
            if os.path.exists("FaceX-Zoo"):
                face_sdk_path = "FaceX-Zoo"
                sys.path.append(face_sdk_path)
        try:
            from core.model_handler.face_detection.FaceDetModelHandler import FaceDetModelHandler
            from core.model_loader.face_detection.FaceDetModelLoader import FaceDetModelLoader
        except ImportError:
            raise ImportError("FaceX-Zoo cannot be imported, please specify the path to the face_sdk path of FaceXZoo"
                              " or put it in the working directory.")

        model_conf = read.yaml(os.path.join(face_sdk_path, "config", "model_conf.yaml"))
        model_path = os.path.join(face_sdk_path, 'models')
        scene = 'non-mask'
        model_category = 'face_detection'
        model_name = model_conf[scene][model_category]

        faceDetModelLoader = FaceDetModelLoader(model_path, model_category, model_name)
        model, cfg = faceDetModelLoader.load_model()
        self.faceDetModelHandler = FaceDetModelHandler(model, device, cfg)

    @staticmethod
    def install(path: Optional[str] = None) -> str:
        """
        Install FaceX-Zoo by clone from GitHub.

        Args:
            path (``str``, optional): The path to install FaceX-Zoo, default is "./FaceX-Zoo".

        Returns:
            ``str``: The path to the installed FaceX-Zoo.

        """
        path = path or "FaceX-Zoo"
        if os.path.exists(path):
            return path

        os.system(f"git clone --depth=1 https://github.com/ControlNet/FaceX-Zoo {path or ''}")
        return path

    def detect_face(self, image: ndarray, *args, **kwargs):
        assert image.ndim == 3 and image.shape[2] == 3, "frame should be 3-dim"
        dets = self.faceDetModelHandler.inference_on_image(image)
        return dets

    def crop_image(self, image_path: str, out_path: str, max_faces=1, margin=0, *args, **kwargs) -> None:
        detector_model_crop_image(self, image_path, out_path, max_faces, margin)

    def crop_video(self, video_path: str, out_path: str, frame_size: Optional[Union[int, Tuple[int, int]]] = None,
        margin=0, fourcc="mp4v", *args, **kwargs
    ) -> None:
        detector_model_crop_video(self, video_path, out_path, frame_size, margin, fourcc)



from functools import partial, update_wrapper, wraps


def _identity_func(arg):
    return arg


class F:
    """
    A function wrapper for pipe operations and partial function.

    Args:
        f (Callable, optional): A function. Default identity function.
        args, kwargs: The arguments to be passed to the function.

    Examples::

        from tensorneko.util import F

        # compose function
        func = F() << (_ + 10) << (_ + 5)
        print(func(10))  # 25

        # the pipe support the tuple as input
        func = F() >> (filter, _ < 6) >> sum
        print(func(range(10))) # 15

        # partial function
        def add(x, y):
            return x + y
        func = F(add, y=1)
        print(func(10)) # 11

    References:
        GitHub - fnpy/fp.py: Missing features of fp in Python -- active fork of kachayev/fp.py. (2022).
        Retrieved 5 April 2022, from https://github.com/fnpy/fn.py

        GitHub - kachayev/fp.py: Functional programming in Python: implementation of missing features to enjoy
        FP. (2022). Retrieved 5 April 2022, from https://github.com/kachayev/fn.py
    """

    def __init__(self, f=_identity_func, *args, **kwargs):
        if args != () or kwargs != {}:
            f = partial(f, *args, **kwargs)
        self.f = f

    @classmethod
    def __compose(cls, f, g):
        return cls(lambda *args, **kwargs: f(g(*args, **kwargs)))

    def __ensure_callable(self, f):
        return self.__class__(*f) if isinstance(f, tuple) else f

    def __rshift__(self, g):
        """Overload >> operator for F instances"""
        return self.__compose(self.__ensure_callable(g), self.f)

    def __lshift__(self, g):
        """Overload << operator for F instances"""
        return self.__compose(self.f, self.__ensure_callable(g))

    def __call__(self, *args, **kwargs):
        """Overload apply operator"""
        return self.f(*args, **kwargs)


def curry(func):
    """
    A decorator that makes the function curried

    Args:
        func (Callable, optional): A function for currying.

    Examples::

        from tensorneko.util.fp import curried

            @curried
            def sum5(a, b, c, d, e):
                return a + b + c + d + e

            print(sum5(1)(2)(3)(4)(5))  # 15
            print(sum5(1, 2, 3)(4, 5))  # 15
    """

    def _args_len(f):
        from inspect import signature
        args = signature(f).parameters
        return len(args)

    @wraps(func)
    def _curried(*args, **kwargs):
        f = func
        count = 0
        while isinstance(f, partial):
            if f.args:
                count += len(f.args)
            f = f.func

        if count == _args_len(f) - len(args):
            return func(*args, **kwargs)

        para_func = partial(func, *args, **kwargs)
        update_wrapper(para_func, f)
        return curry(para_func)

    return _curried



from typing import Iterable, Optional

from ..backend.visual_lib import VisualLib
from ..io._default_backends import _default_video_io_backend


def frames2video(frame_paths: Iterable[str], output_path: str, fps: int, backend: Optional[VisualLib] = None) -> None:
    """
    Convert frames to video.

    Args:
        frame_paths (``Iterable[str]``): Paths of frames.
        output_path (``str``): Path of output video.
        fps (``int``): Frames per second.
        backend (:class:`~tensorneko_util.backend.visual_lib.VisualLib`, optional): The backend library for saving.
            Default: OPENCV if installed else PYTORCH then FFMPEG.
    """
    backend = backend or _default_video_io_backend()

    if backend == VisualLib.OPENCV:
        if not VisualLib.opencv_available():
            raise ValueError("OpenCV is not installed.")

        import cv2
        ext = output_path.split('.')[-1]
        if ext == 'avi':
            fourcc = cv2.VideoWriter_fourcc(*'XVID')
        elif ext == 'mp4':
            fourcc = cv2.VideoWriter_fourcc(*'MP4V')
        else:
            raise ValueError("Video ext only supports 'avi' or 'mp4'.")

        video = None

        for frame_path in frame_paths:
            img = cv2.imread(frame_path)
            if video is None:
                video = cv2.VideoWriter(output_path, fourcc, fps, img.shape[1::-1])
            video.write(img)

        video.release()

    elif backend == VisualLib.PYTORCH:
        if not VisualLib.pytorch_available():
            raise ValueError("Torchvision is not installed.")

        import torchvision
        import torch

        frames = []

        for frame_path in frame_paths:
            frames.append(torchvision.io.read_image(frame_path))

        torchvision.io.write_video(output_path, torch.stack(frames, dim=0).permute(0, 2, 3, 1), fps=fps)

    elif backend == VisualLib.FFMPEG:
        raise NotImplementedError("FFMPEG is not implemented yet.")

    else:
        raise ValueError("Supported backend is only OpenCV and PyTorch.")



import pathlib
from typing import Optional

import numpy as np
from einops import rearrange
from numpy import ndarray

from .._default_backends import _default_image_io_backend
from ...backend.visual_lib import VisualLib
from ...util.type import T_ARRAY


class ImageWriter:
    """ImageWriter for saving images from :class:`~torch.Tensor` or :class:`~numpy.ndarray`"""

    @classmethod
    def _convert_img_format(cls, img_in: T_ARRAY, backend: VisualLib, channel_first: bool):
        # torchvision requires (C, H, W) but others requires (H, W, C), normalize the shape
        if backend == VisualLib.PYTORCH and not channel_first:
            img_in = rearrange(img_in, 'H W C -> C H W')
        elif backend != VisualLib.PYTORCH and channel_first:
            img_in = rearrange(img_in, 'C H W -> H W C')

        # convert value range to [0, 255] int
        if isinstance(img_in, ndarray):
            img_out = (img_in * 255).astype(np.uint8)
        else:
            try:
                import torch
                if isinstance(img_in, torch.Tensor):
                    img_out = (img_in * 255).type(torch.IntTensor)
                else:
                    raise ValueError("Unknown data type. The image array type must be numpy.ndarray or torch.Tensor.")
            except ImportError:
                raise ValueError("Unknown data type. The image array type must be numpy.ndarray or torch.Tensor.")

        return img_out

    @classmethod
    def to_jpeg(cls, path: str, image: T_ARRAY, quality: int = 75, channel_first: bool = False,
        backend: Optional[VisualLib] = None
    ) -> None:
        """
        Save as jpeg files from :class:`~torch.Tensor` or :class:`~numpy.ndarray`. The value range is [0, 1].

        Args:
            path (``str``): The path of output file.
            image (:class:`~torch.Tensor` | :class:`~numpy.ndarray`): The image tensor for output.
            quality (``int``, optional): The quality level for jpg from 0 to 100. Higher means quality is better.
                Default: 75.
            channel_first (``bool``, optional): The flag for channel first.
                Specify the image shape is (H, W, C) if False, else (C, H, W) if True Default: False.
            backend (:class:`~tensorneko_util.backend.visual_lib.VisualLib`, optional): The backend library for saving.
                Default: "opencv" if installed else "matplotlib".
        """
        backend = backend or _default_image_io_backend()
        image = cls._convert_img_format(image, backend, channel_first)

        if backend == VisualLib.OPENCV:
            import cv2
            image = cv2.cvtColor(image, cv2.COLOR_RGB2BGR)
            cv2.imwrite(path, image, [cv2.IMWRITE_JPEG_QUALITY, quality])
        elif backend == VisualLib.MATPLOTLIB:
            import matplotlib.pyplot as plt
            plt.imsave(path, image, format='jpeg')
        elif backend == VisualLib.PIL:
            import PIL.Image
            PIL.Image.fromarray(image).save(path, format='jpeg', quality=quality)
        elif backend == VisualLib.PYTORCH:
            import torchvision.io
            torchvision.io.write_jpeg(image, path, quality=quality)
        else:
            raise ValueError("Unknown backend library.")

    @classmethod
    def to_png(cls, path: str, image: T_ARRAY, compression_level: int = 6, channel_first: bool = False,
        backend: Optional[VisualLib] = None
    ) -> None:
        """
        Save as png files from :class:`~torch.Tensor` or :class:`~numpy.ndarray` with (C, H, W).

        Args:
            path (``str``): The path of output file.
            image (:class:`~torch.Tensor` | :class:`~numpy.ndarray`): The image tensor for output.
            compression_level (``int``, optional): The compression level png from 0 to 9. Higher means quality is worse.
                Default: 6
            channel_first (``bool``, optional): The flag for channel first.
                Specify the image shape is (H, W, C) if False, else (C, H, W) if True Default: False.
            backend (:class:`~tensorneko_util.backend.visual_lib.VisualLib`, optional): The backend library for saving.
                Default: "opencv" if installed else "matplotlib".
        """
        backend = backend or _default_image_io_backend()
        image = cls._convert_img_format(image, backend, channel_first)
        if backend == VisualLib.OPENCV:
            if not VisualLib.opencv_available():
                raise ValueError("OpenCV is not installed.")
            import cv2
            image = cv2.cvtColor(image, cv2.COLOR_RGB2BGR)
            cv2.imwrite(path, image, [cv2.IMWRITE_PNG_COMPRESSION, compression_level])
        elif backend == VisualLib.MATPLOTLIB:
            if not VisualLib.matplotlib_available():
                raise ValueError("Matplotlib is not installed.")
            import matplotlib.pyplot as plt
            plt.imsave(path, image, format='png')
        elif backend == VisualLib.PIL:
            if not VisualLib.pil_available():
                raise ValueError("PIL is not installed.")
            import PIL.Image
            PIL.Image.fromarray(image).save(path, format='png', compress_level=compression_level)
        elif backend == VisualLib.PYTORCH:
            if not VisualLib.pytorch_available():
                raise ValueError("Torchvision is not installed.")
            import torchvision.io
            torchvision.io.write_png(image, path, compression_level=compression_level)
        else:
            raise ValueError("Unknown backend library.")

    @classmethod
    def to(cls, path: str, image: T_ARRAY, *args, **kwargs) -> None:
        """
        Save as png files from :class:`~torch.Tensor` or :class:`~numpy.ndarray` with (C, H, W).

        Args:
            path (``str``): The path of output file.
            image (:class:`~torch.Tensor` | :class:`~numpy.ndarray`): The image tensor for output.
            *args: The arguments for :meth:`ImageWriter.to_jpeg` or :meth:`ImageWriter.to_png`.
            **kwargs: The keyword arguments for :meth:`ImageWriter.to_jpeg` or :meth:`ImageWriter.to_png`.
        """
        ext = pathlib.Path(path).suffix
        if ext in ('.jpg', '.jpeg'):
            cls.to_jpeg(path, image, *args, **kwargs)
        elif ext == '.png':
            cls.to_png(path, image, *args, **kwargs)
        else:
            raise ValueError("Unknown file extension. Now only support .jpg, .jpeg and .png.")

    def __new__(cls, path: str, image: T_ARRAY, *args, **kwargs) -> None:
        """Alias of :meth:`ImageWriter.to`"""
        cls.to(path, image, *args, **kwargs)



from __future__ import annotations

import json
import os
import shutil
import glob
from typing import Collection

from ...util import tensorneko_util_path
from ...util.server import HttpThreadServer


class Server(HttpThreadServer):

    def __init__(self, path: str, port: int = 8000, exts: Collection[str] = ("png", "jpg", "jpeg")):
        super().__init__(path, port, "Tensorneko Image Browser")
        self.exts = exts

    def _prepare(self) -> None:
        # write a metadata json file list oll images in that directory
        metadata = []
        for ext in self.exts:
            metadata.extend(glob.glob(os.path.join(self.server_dir, "**", f"*.{ext}"), recursive=True))

        metadata.sort()
        metadata = [os.path.relpath(x, self.server_dir) for x in metadata]

        with open(os.path.join(self.server_dir, ".metadata.json"), "w") as f:
            json.dump(sorted(metadata), f, indent=4, ensure_ascii=True)

        # copy index.html to target directory
        shutil.copyfile(
            os.path.join(tensorneko_util_path, "visualization", "image_browser", "web", "index.html"),
            os.path.join(self.server_dir, "index.html")
        )

    def stop(self) -> None:
        super().stop()
        # remove the metadata file
        os.remove(os.path.join(self.server_dir, ".metadata.json"))
        # remove the index.html file
        os.remove(os.path.join(self.server_dir, "index.html"))



from typing import List

import numpy as np
from numpy import ndarray

from .dispatcher import dispatch


@dispatch
def sparse2binary(x: ndarray, length: int = None) -> ndarray:
    """
    Convert sparse vector to binary vector.

    Args:
        x (:class:`numpy.ndarray`): The sparse vector.
        length (``int``, optional): The length of binary vector.

    Returns:
        :class:`numpy.ndarray`: The binary vector.

    Examples::

        x = np.array([1, 2])
        sparse2binary(x, length=5)
        array([0, 1, 1, 0, 0])
    """
    if length is None:
        length = x.max() + 1
    binary = np.zeros(length, dtype=np.int32)
    binary[x] = 1
    return binary


@dispatch.of(list)
def sparse2binary(x: List[int]) -> List[int]:
    return sparse2binary(np.array(x)).tolist()


@dispatch.of(list, int)
def sparse2binary(x: List[int], length: int) -> List[int]:
    return sparse2binary(np.array(x), length=length).tolist()


@dispatch
def binary2sparse(x: ndarray) -> ndarray:
    """
    Convert binary vector to sparse vector.

    Args:
        x (:class:`numpy.ndarray`): The binary vector.

    Returns:
        :class:`numpy.ndarray`: The sparse vector.

    Examples::

        x = np.array([0, 1, 1, 0, 0])
        binary2sparse(x, length=5)
        array([1, 2])
    """
    return np.where(x == 1)[0]


@dispatch.of(list)
def binary2sparse(x: List[int]) -> List[int]:
    return [i for i, v in enumerate(x) if v == 1]



from __future__ import annotations

from dataclasses import dataclass
from typing import Tuple, Optional, Any, List, Union, Generic, Callable
from .type import T


@dataclass
class WindowValue(Generic[T]):
    start: float
    end: float
    value: T


@dataclass
class WindowValueCache(Generic[T]):
    start: float
    end: float
    value: List[T]

    def merge_value(self, merge_func: Callable[[List[T]], T]) -> WindowValue[T]:
        return WindowValue(self.start, self.end, merge_func(self.value))

    def __eq__(self, other: WindowValueCache[T]):
        return self.start == other.start and self.end == other.end


class WindowMerger:

    def __init__(self):
        self._data: List[WindowValueCache] = []

    @classmethod
    def _get_overlap(cls, prev_start: float, prev_end: float, new_start: float, new_end: float
    ) -> Tuple[int, Tuple[Optional[float], Optional[float]]]:
        """
        Get overlap of two windows. And consider 9 cases:
        1. prev_start < new_start < prev_end < new_end
            prev: |----------|
            new :     |----------|
        2. new_start < prev_start < new_end < prev_end
            prev:     |----------|
            new : |----------|
        3. prev_start < new_start < new_end < prev_end
            prev: |--------------|
            new :     |------|
        4. new_start < prev_start < prev_end < new_end
            prev:     |------|
            new : |--------------|
        5. new_start == prev_start and new_end == prev_end
            prev: |----------|
            new : |----------|
        6. new_start == prev_start and new_end < prev_end
            prev: |----------|
            new : |------|
        7. new_start == prev_start and new_end > prev_end
            prev: |------|
            new : |----------|
        8. new_start > prev_start and new_end == prev_end
            prev: |----------|
            new :     |------|
        9. new_start < prev_start and new_end == prev_end
            prev:     |------|
            new : |----------|

        Args:
            prev_start: start of previous window
            prev_end: end of previous window
            new_start: start of new window
            new_end: end of new window

        Returns:
            (overlap_case, (overlap_start, overlap_end)):
                overlap_case is 1, 2, 3 or 4.
                overlap_start and overlap_end is the start and end of overlap window.
        """
        if prev_start < new_start < prev_end < new_end:
            return 1, (new_start, prev_end)
        elif new_start < prev_start < new_end < prev_end:
            return 2, (prev_start, new_end)
        elif prev_start < new_start < new_end < prev_end:
            return 3, (new_start, new_end)
        elif new_start < prev_start < prev_end < new_end:
            return 4, (prev_start, prev_end)
        elif new_start == prev_start and new_end == prev_end:
            return 5, (new_start, new_end)
        elif new_start == prev_start and new_end < prev_end:
            return 6, (new_start, new_end)
        elif new_start == prev_start and new_end > prev_end:
            return 7, (new_start, prev_end)
        elif new_start > prev_start and new_end == prev_end:
            return 8, (new_start, new_end)
        elif new_start < prev_start and new_end == prev_end:
            return 9, (prev_start, new_end)
        else:
            return 0, (None, None)

    @classmethod
    def _merge_step(cls, old_item: WindowValueCache, new_item: WindowValueCache
    ) -> Tuple[List[WindowValueCache], List[WindowValueCache], int]:
        start, end, value = new_item.start, new_item.end, new_item.value
        merged_items = []
        new_items = []

        overlap_case, overlap = cls._get_overlap(old_item.start, old_item.end, start, end)
        if overlap_case == 0:
            # No overlap
            merged_items.append(old_item)
            new_items.append(new_item)
        elif overlap_case == 1:
            # prev: |----------|
            # new :     |++++++++++|
            #       |---|------|+++|
            merged_items.append(WindowValueCache(old_item.start, overlap[0], old_item.value))
            merged_items.append(WindowValueCache(overlap[0], overlap[1], old_item.value + value))
            new_items.append(WindowValueCache(overlap[1], end, value))
        elif overlap_case == 2:
            # prev:     |----------|
            # new : |++++++++++|
            #       |+++|------|---|
            new_items.append(WindowValueCache(start, overlap[0], value))
            merged_items.append(WindowValueCache(overlap[0], overlap[1], old_item.value + value))
            merged_items.append(WindowValueCache(overlap[1], old_item.end, old_item.value))
        elif overlap_case == 3:
            # prev: |--------------|
            # new :     |++++++|
            #       |---|------|---|
            merged_items.append(WindowValueCache(old_item.start, overlap[0], old_item.value))
            merged_items.append(WindowValueCache(overlap[0], overlap[1], old_item.value + value))
            merged_items.append(WindowValueCache(overlap[1], old_item.end, old_item.value))
        elif overlap_case == 4:
            # prev:     |------|
            # new : |++++++++++++++|
            #       |+++|------|+++|
            new_items.append(WindowValueCache(start, overlap[0], value))
            merged_items.append(WindowValueCache(overlap[0], overlap[1], old_item.value + value))
            new_items.append(WindowValueCache(overlap[1], end, value))
        elif overlap_case == 5:
            # prev: |--------------|
            # new : |++++++++++++++|
            #       |--------------|
            merged_items.append(WindowValueCache(old_item.start, old_item.end, old_item.value + value))
        elif overlap_case == 6:
            # prev: |--------------|
            # new : |++++++|
            #       |------|-------|
            merged_items.append(WindowValueCache(old_item.start, overlap[1], old_item.value + value))
            merged_items.append(WindowValueCache(overlap[1], old_item.end, old_item.value))
        elif overlap_case == 7:
            # prev: |------|
            # new : |++++++++++++++|
            #       |------|+++++++|
            merged_items.append(WindowValueCache(old_item.start, old_item.end, old_item.value + value))
            new_items.append(WindowValueCache(overlap[1], end, value))
        elif overlap_case == 8:
            # prev: |--------------|
            # new :       |++++++++|
            #       |-----|--------|
            merged_items.append(WindowValueCache(old_item.start, overlap[0], old_item.value))
            merged_items.append(WindowValueCache(overlap[0], old_item.end, old_item.value + value))
        elif overlap_case == 9:
            # prev:         |-----|
            # new : |+++++++++++++|
            #       |+++++++|-----|
            new_items.append(WindowValueCache(start, overlap[0], value))
            merged_items.append(WindowValueCache(overlap[0], old_item.end, old_item.value + value))

        return merged_items, new_items, overlap_case

    def add(self, start: float, end: float, value: Any) -> None:
        if end <= start:
            raise ValueError("end must be larger than start")

        if len(self._data) == 0:
            self._data.append(WindowValueCache(start, end, [value]))
            return

        old_data = self._data.copy()
        old_data_overlapped = []

        new_data = []
        # find all overlap windows
        for old_item in old_data:
            overlap_case, _ = self._get_overlap(old_item.start, old_item.end, start, end)
            if overlap_case == 0:
                # if it is not overlapped, keep the old item
                new_data.append(old_item)
            else:
                # if it is overlapped, record the overlapped old items
                old_data_overlapped.append(old_item)

        # merge the overlapped old items
        new_items = [WindowValueCache(start, end, [value])]
        while len(old_data_overlapped) > 0:
            old_item = old_data_overlapped.pop(0)
            new_items_step = []

            old_data_is_overlap = False
            for new_item in new_items:
                merged_items, _new_items, overlap_case = self._merge_step(old_item, new_item)
                if overlap_case != 0:
                    old_data_is_overlap = True

                if overlap_case in (4, 5, 7, 9):
                    # if the old item is wrapped in the new item, append the old item to the result
                    new_data.extend(merged_items)
                else:
                    # if the old item is not wrapped in the new item, only append if it's not overlapped
                    new_data.extend([item for item in merged_items if item != old_item])
                new_items_step.extend(_new_items)

            if not old_data_is_overlap:
                new_data.append(old_item)

            new_items = new_items_step
            if len(new_items) == 0:
                break
        else:
            new_data.extend(new_items)

        self._data = sorted(new_data, key=lambda x: x.start)

    def merge(self) -> List[WindowValue]:
        return [each.merge_value(self.number_merge) for each in self._data]

    @staticmethod
    def number_merge(values: List[Union[int, float]]) -> float:
        return sum(values) / len(values)



from typing import Optional

import numpy as np
from numpy import ndarray

from ..backend import VisualLib


def rgb2gray(img: ndarray, channel_first: bool = False, backend: Optional[VisualLib] = VisualLib.PIL) -> ndarray:
    """
    Convert RGB image to gray scale image.

    Args:
        img (:class:`np.ndarray`): RGB image. (H, W, 3)
        channel_first (``bool``, optional): If ``True``, the image is in channel first format. (3, H, W)
        backend (:class:`~tensorneko_util.backend.visual_lib.VisualLib`, optional): The backend library for saving. Now
            supports PIL, OpenCV and skimage. Default: PIL.

    Returns:
        :class:`np.ndarray`: Gray scale image. (H, W)
    """
    if channel_first:
        img = img.transpose(1, 2, 0)

    if backend == VisualLib.PIL:
        if not VisualLib.pil_available():
            raise ValueError("Pillow is not installed.")
        from PIL import Image
        # should ensure the format is uint8
        if img.max() <= 1:
            img = (img * 255).astype(np.uint8)
        return np.asarray(Image.fromarray(img).convert('L')) / 255

    elif backend == VisualLib.OPENCV:
        if not VisualLib.opencv_available():
            raise ValueError("OpenCV is not installed.")
        import cv2
        # should ensure the format is uint8
        if img.max() <= 1:
            img = (img * 255).astype(np.uint8)
        return cv2.cvtColor(img, cv2.COLOR_RGB2GRAY) / 255

    elif backend == VisualLib.SKIMAGE:
        if not VisualLib.skimage_available():
            raise ValueError("scikit-image is not installed.")
        import skimage.color
        if img.dtype == np.uint8 or img.max() > 1:
            img = img / 255
        return skimage.color.rgb2gray(img)

    else:
        raise ValueError(f"Backend {backend} is not supported.")


def rgb2gray_batch(img: ndarray, channel_first: bool = False, backend: Optional[VisualLib] = VisualLib.PIL) -> ndarray:
    """
    Convert RGB images to gray scale images.

    Args:
        img (:class:`np.ndarray`): RGB images. (N, H, W, 3)
        channel_first (``bool``, optional): If ``True``, the image is in channel first format. (N, 3, H, W)
        backend (:class:`~tensorneko_util.backend.visual_lib.VisualLib`, optional): The backend library for saving. Now
            supports PIL, OpenCV and skimage. Default: PIL.

    Returns:
        :class:`np.ndarray`: Gray scale images. (N, H, W)
    """
    if channel_first:
        img = img.transpose(0, 2, 3, 1)

    return np.asarray([rgb2gray(i, channel_first=False, backend=backend) for i in img])
